# import pytest

from tral.sequence import repeat_detection_run, sequence

TEST_SEQUENCE_Q9BRR0 = "MARELSESTALDAQSTEDQMELLVIKVEEEEAGFPSSPDLGSEGSRERFRGFRYPEAAGPREALSRLRELCRQWLQPEMHSKEQILELLVLEQFLTILPGNLQSWVREQHPESGEEVVVLLEYLERQLDEPAPQVSGVDQGQELLCCKMALLTPAPGSQSSQFQLMKALLKHESVGSQPLQDRVLQVPVLAHGGCCREDKVVASRLTPESQGLLKVEDVALTLTPEWTQQDSSQGNLCRDEKQENHGSLVSLGDEKQTKSRDLPPAEELPEKEHGKISCHLREDIAQIPTCAEAGEQEGRLQRKQKNATGGRRHICHECGKSFAQSSGLSKHRRIHTGEKPYECEECGKAFIGSSALVIHQRVHTGEKPYECEECGKAFSHSSDLIKHQRTHTGEKPYECDDCGKTFSQSCSLLEHHRIHTGEKPYQCSMCGKAFRRSSHLLRHQRIHTGDKNVQEPEQGEAWKSRMESQLENVETPMSYKCNECERSFTQNTGLIEHQKIHTGEKPYQCNACGKGFTRISYLVQHQRSHVGKNILSQ"

# @pytest.mark.slow


def test_detect_TRUST():

    test_seq = sequence.Sequence(TEST_SEQUENCE_Q9BRR0)
    predicted_repeats = repeat_detection_run.run_detector(seq_records=[test_seq], detectors=["TRUST"])[0]['TRUST']
    # Warning: TRUST finds these results ONLY with the BLOSUM50 substitution matrix.
    assert len(predicted_repeats) == 3
    # Warning: TRUST finds these results ONLY with the BLOSUM50 substitution matrix.
    assert predicted_repeats[0].msa == ['HLREDIAQIP---TCAEAGE---QEGRLQR', 'KQKNATGGRR--HICHECGKSFAQSSGLSK', 'HRRIHTGEKP--YECEECGKAFIGSSALVI', 'HQRVHTGEKP--YECEECGKAFSHSSDLIK', 'HQRTHTGEKP--YECDDCGKTFSQSCSLLE', 'HHRIHTGEKP--YQCSMCGKAFRRSSHLLR', 'HQRIHTGDKN--VQEPEQGEAW--KSRM--', 'ESQLENVETPmsYKCNECERSFTQNTGLIE', 'HQKIHTGEKP--YQCNACGKGFTRISYLVQ']


def test_detect_XSTREAM():
    test_seq = sequence.Sequence(TEST_SEQUENCE_Q9BRR0)
    predicted_repeats = repeat_detection_run.run_detector(seq_records=[test_seq], detectors=["XSTREAM"])[0]['XSTREAM']
    assert len(predicted_repeats) == 1
    assert predicted_repeats[0].msa == ['ECGKSFAQS-SGLSK-HRRIHTGEKPYECE', 'ECGKAFIGS-SALVI-HQRVHTGEKPYECE', 'ECGKAFSHS-SDL-IKHQRTHTGEKPYECD', 'DCGKTFSQSCSLLEH-H-RIHTGEKPY']


def test_detect_TREKS():

    test_seq = sequence.Sequence(TEST_SEQUENCE_Q9BRR0)
    predicted_repeats = repeat_detection_run.run_detector(seq_records=[test_seq], detectors=["T-REKS"])[0]['T-REKS']
    assert len(predicted_repeats) == 1
    assert predicted_repeats[0].msa == ['C---G---KSFAQSSGLSKHRRIHTGEKPYECE-E', 'C---G---KAFIGSSALVIHQRVHTGEKPYECE-E', 'C---G---KAFSHSSDLIKHQRTHTGEKPYECD-D', 'C---G---KTFSQSCSLLEHHRIHTGEKPYQCS-M', 'C---G---KAFRRSSHLLRHQRIHTGDKNVQ-EPE', 'Q---G---EAW--KSRME-SQ-LENVETPMSYK--', 'C---NECERSFTQNTGLIEHQKIHTGEKPYQ----', 'CNACG---KGFTRISYLVQHQRSHVG-KNI-LS--']


def test_detect_HHrepID():

    test_seq = sequence.Sequence(TEST_SEQUENCE_Q9BRR0)
    predicted_repeats = repeat_detection_run.run_detector(seq_records=[test_seq], detectors=["HHrepID"])[0]['HHrepID']
    assert len(predicted_repeats) == 1
    expected_msa = ['------------------IPTCAEAGEQ----',
                    'EGRLQRKQKNATGGRRHICHECGKSFAQ----',
                    'SSGLSKHRRIHTGEKPYECEECGKAFIG----',
                    'SSALVIHQRVHTGEKPYECEECGKAFSH----',
                    'SSDLIKHQRTHTGEKPYECDDCGKTFSQ----',
                    'SCSLLEHHRIHTGEKPYQCSMCGKAFRR----',
                    'SSHLLRHQRIHTGDKNVQEPEQGEAWKSRMES',
                    '------QLENVETPMSYKCNECERSFTQ----',
                    'NTGLIEHQKIHTGEKPYQCNACGKGFTR----',
                    'ISYLVQHQRSHVG-------------------']
    assert predicted_repeats[0].msa == expected_msa



#! /usr/bin/env python

import configobj
import os
import os.path
import shlex
import shutil
import sqlalchemy as sqla

# Interface to Gc3libs

import gc3libs
from gc3libs import Application, Run
from gc3libs.cmdline import SessionBasedScript
from gc3libs.workflow import SequentialTaskCollection, ParallelTaskCollection
from gc3libs.persistence.accessors import GetValue
from gc3libs.quantity import MB
import gc3libs.debug
import gc3libs.utils


######################## Basic Applications/Tasks Templates ##############

class MyApplication(Application):

    """
    Basic template method pattern  `Application`:class: Initialise Application generically,
    and check for successful running generically.
    """
    # @gc3libs.debug.trace

    def __init__(self, name, **kwargs):

        gc3libs.log.info("Initialising {}".format(self.__class__.__name__))
        config = kwargs["config"]
        self.c = config[name].copy()
        self.jobname = self.__class__.__name__  # Used by GC3PIE

        kwargs['output_dir'] = self.c['logdir']

        # Replace every "%X" in the config with the current value for X, e.g.
        # "3".
        if "param" in kwargs:
            for iC in self.c.keys():
                for param_name, param_value in kwargs['param'].items():
                    self.c[iC] = self.c[iC].replace(param_name, param_value)

            for param_name, param_value in kwargs['param'].items():
                if param_name == '$N':
                    self.N = param_value
                # Adapt output_dir to particular file
                kwargs['output_dir'] = kwargs['output_dir'].replace(
                    param_name,
                    param_value)

        if "required_memory" in self.c:
            kwargs['requested_memory'] = int(self.c['required_memory']) * MB
        else:
            kwargs['requested_memory'] = int(config['required_memory']) * MB

        gc3libs.Application.__init__(
            self,
            arguments=shlex.split(
                self.c['script']) +
            ["-i"] +
            shlex.split(
                self.c['input']) +
            ["-o"] +
            shlex.split(
                self.c['output']) +
            shlex.split(
                self.c['extra']),
            inputs=[],
            outputs=[],
            join=True,
            stdout=self.c['stdout'],
            stderr=self.c['stderr'],
            **kwargs)

    def terminated(self):
        gc3libs.log.info(
            "Testing whether {} is done".format(
                self.__class__.__name__))

        # If the application has terminated o.k. (self.execution.returncode == 0),
        #   Check wether all resultfiles are o.k. If yes: Good, If no: FREEZE.
        # If not: Save error and FREEZE.

        if self.execution.returncode == 0:
            gc3libs.log.info(
                "{1} claims to be successful: self.execution.returncode: {0}".format(
                    self.execution.returncode,
                    self.__class__.__name__))
            # Check if result file exists (Later: and complies to some rules).
            if self.is_valid(self.c['output']):  # IMPLEMENT IN CLASS!
                gc3libs.log.info(
                    "{} has run successfully to completion.".format(
                        self.__class__.__name__))
                # Now, clean up
                # Delete all log files if everything ran smoothly.
                shutil.rmtree(self.output_dir)
            else:
                gc3libs.log.info(
                    "%s has not produced a valid outputfile.",
                    self.__class__.__name__)
                # Set self.execution.exitcode to a non-zero integer <256 (to
                # indicate an error)
                self.execution.exitcode = 42

        else:
            gc3libs.log.info(
                "{1} is not successful: self.execution.returncode: {0}".format(
                    self.execution.returncode,
                    self.__class__.__name__))
            # Check if there is stderr.
            if not os.path.isfile(os.path.join(self.output_dir, self.stderr)):
                self.error_tag = ""
            else:
                # Create a tag from the last line in stderr.
                with open(os.path.join(self.output_dir, self.stderr), "r") as fh:
                    for line in fh:
                        pass
                    self.error_tag = line
            self.execution.exitcode = 88


class StopOnError(object):

    """
    Mix-in class to make a `SequentialTaskCollection`:class: turn to STOPPED
    state as soon as one of the tasks fail.
    """

    def next(self, done):
        if done == len(self.tasks) - 1:
            self.execution.returncode = self.tasks[done].execution.returncode
            return Run.State.TERMINATED
        else:
            rc = self.tasks[done].execution.exitcode
            if rc != 0:
                return Run.State.STOPPED
            else:
                return Run.State.RUNNING


################################# Applications/Tasks #####################

class FilePreparation(MyApplication):

    def is_valid(self, output):

        #        if not os.path.isfile(self.c['output']):
        #        gc3libs.log.info("{} has not produced an output file.".format(self.__class__.__name__))
        #        self.execution.returncode = 1
        #        # Set self.execution.exitcode to a non-zero integer <256 (to indicate an error)
        #        self.execution.exitcode = 42

        return True


class AnnotateTandemRepeats(MyApplication):

    def is_valid(self, output):
        return True


class SerializeAnnotations(MyApplication):

    def is_valid(self, output):
        return True


############################# Main Session Creator (?) ###################


class TandemRepeatAnnotationWorkflow(SessionBasedScript):

    """
    TandemRepeatAnnotationWorkflow
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version='0.0.1',
        )

    def setup_options(self):
        self.add_param("-conf", "--config_file", type=str, required=True,
                       help="Path to the parameter file")

    def _make_session(self, session_uri, store_url):
        return gc3libs.session.Session(
            session_uri,
            store_url,
            extra_fields={
                # NB: enlarge window to at least 150 columns to read this table
                # properly!
                # Former: Class                                           , #
                # task class
                sqla.Column('jobname', sqla.TEXT()): (lambda obj: obj.__class__.__name__),
                # .ONLY(CodemlApplication), # program executable
                sqla.Column('executable', sqla.TEXT()): GetValue(default=None) .arguments[0],
                # .ONLY(CodemlApplication), # fullpath to codeml output directory
                sqla.Column('output_path', sqla.TEXT()): GetValue(default=None) .output_dir,
                sqla.Column('cluster', sqla.TEXT()): GetValue(default=None) .execution.resource_name,  # .ONLY(CodemlApplication), # cluster/compute element
                # sqla.Column('worker',             sqla.TEXT())    : GetValue(default=None) .hostname                          ,#.ONLY(CodemlApplication), # hostname of the worker node
                # sqla.Column('cpu',                sqla.TEXT())    : GetValue(default=None) .cpuinfo                           ,#.ONLY(CodemlApplication), # CPU model of the worker node
                # sqla.Column('requested_walltime', sqla.INTEGER()) : _get_requested_walltime_or_none                           , # requested walltime, in hours
                sqla.Column('requested_cores', sqla.INTEGER()): GetValue(default=None) .requested_cores,  # .ONLY(CodemlApplication), # num of cores requested
                sqla.Column('duration_s', sqla.INTEGER()): lambda app: app.execution.duration.amount(unit=app.execution.duration.second),  # .ONLY(CodemlApplication), # used walltime
                sqla.Column('max_used_memory_mb', sqla.INTEGER()): lambda app: app.execution.max_used_memory.amount(unit=app.execution.max_used_memory.MB),  # .ONLY(CodemlApplication), # used max_used_memory # might need to be saved as TEXT
                sqla.Column('lrms_jobid', sqla.TEXT()): GetValue(default=None) .execution.lrms_jobid,  # .ONLY(CodemlApplication), # arc job ID
                # sqla.Column('original_exitcode',  sqla.INTEGER()) : GetValue(default=None) .execution.original_exitcode       ,#.ONLY(CodemlApplication), # original exitcode
                sqla.Column('used_cpu_time_s', sqla.INTEGER()): lambda app: app.execution.used_cpu_time.amount(unit=app.execution.used_cpu_time.second),  # .ONLY(CodemlApplication), # used walltime
                # returncode = exitcode*256 + signal
                sqla.Column('returncode', sqla.INTEGER()): GetValue(default=None) .execution.returncode,  # .ONLY(CodemlApplication), # returncode attr
                # sqla.Column('queue',              sqla.TEXT())    : GetValue(default=None) .execution.queue                   ,#.ONLY(CodemlApplication), # exec queue _name_
                sqla.Column('time_submitted', sqla.FLOAT()): GetValue(default=None) .execution.timestamp['SUBMITTED'],  # .ONLY(CodemlApplication), # client-side submission (float) time
                sqla.Column('time_terminated', sqla.FLOAT()): GetValue(default=None) .execution.timestamp['TERMINATED'],  # .ONLY(CodemlApplication), # client-side termination (float) time
                sqla.Column('time_stopped', sqla.FLOAT()): GetValue(default=None) .execution.timestamp['STOPPED'],  # .ONLY(CodemlApplication), # client-side stop (float) time
                sqla.Column('error_tag', sqla.TEXT()): GetValue(default=None) .error_tag,
                sqla.Column('N', sqla.TEXT()): GetValue(default=None).N.ONLY((tandem_repeat_annotation_workflow.AnnotateTandemRepeats)),
            })

    def parse_args(self):
        self.config_file = self.params.config_file
        gc3libs.log.info(
            "TestWorkflow config_file: {}".format(
                self.config_file))

    def new_tasks(self, kwargs):

        #name = "myTestWorkflow"
        gc3libs.log.info("Calling TestWorkflow.next_tasks()")

        self.config = configobj.ConfigObj(self.config_file, stringify=True)

        yield tandem_repeat_annotation_workflow.MainSequentialFlow(config=self.config, **kwargs)


######################## Support Classes / Workflow elements #############


class MainSequentialFlow(SequentialTaskCollection):

    def __init__(self, **kwargs):

        config = kwargs["config"]
        self.kwargs = kwargs
        gc3libs.log.info(
            "\t Calling MainSequentialFlow.__init({})".format("<No parameters>"))

        self.initial_tasks = []
        if config["file_preparation"]["activated"] == 'True':
            self.initial_tasks = [
                FilePreparation(
                    name="file_preparation",
                    **kwargs)]
        else:
            self.initial_tasks = [SequencewiseParallelFlow(**kwargs)]

        SequentialTaskCollection.__init__(self, self.initial_tasks, **kwargs)

    def next(self, done):
        last_task = self.tasks[done]
        rc = last_task.execution.exitcode
        gc3libs.log.info(
            "\t MainSequentialFlow: Find next task. last_task.execution.exitcode: [%s]" %
            rc)
        # Stop the execution if last application failed.
        if rc != 0:
            gc3libs.log.info("\t rc is not 0 [%s]" % type(rc))
            self.execution.rc = rc
            return Run.State.STOPPED

        # Shall we continue with the workflow?
        if last_task != self.tasks[-1]:
            gc3libs.log.info("\t MainSequentialFlow: Do not add anything")
            # We still have tasks to run in self.tasks, let's consume them
            # before adding new tasks.
            return Run.State.RUNNING
        elif isinstance(last_task, DataPreparationParallelFlow) or isinstance(last_task, FilePreparation):
            gc3libs.log.info(
                "\t MainSequentialFlow: Add SequencewiseParallelFlow")
            # SequencewiseParallelFlow task needs to be initialized using the
            # output from previous steps.
            self.add(SequencewiseParallelFlow(**self.kwargs))
            return Run.State.RUNNING
        else:
            gc3libs.log.info("\t MainSequentialFlow: Terminated")
            # Workflow is terminated.
            self.execution.rc = rc
            return Run.State.TERMINATED

    def terminated(self):
        self.execution.returncode = 0
        gc3libs.log.info(
            "\t MainSequentialFlow.terminated [%s]" %
            self.execution.returncode)


class DataPreparationParallelFlow(ParallelTaskCollection):

    def __init__(self, **kwargs):

        self.kwargs = kwargs
        gc3libs.log.info(
            "\t\tCalling DataPreparationParallelFlow.__init({})".format(
                self.kwargs))

        # Needs to be coded


class SequencewiseParallelFlow(ParallelTaskCollection):

    def __init__(self, **kwargs):

        config = kwargs["config"]
        self.c = config["sequencewise_parallel_flow"]

        # TODO: Find all files in dir and create self.lSeq! Warning! Should be done once the
        # Tasks before are finished.
        #self.lSeq = [re.findall(self.c['retag'], i)[0] for i in os.listdir(self.c['input'])]
        self.lSeq = [
            i for i in os.listdir(
                self.c['input']) if not i.endswith("fai")]
        self.kwargs = kwargs

        gc3libs.log.info(
            "\t\tCalling SequencewiseParallelFlow.__init({})".format(
                self.kwargs))

        self.tasks = [
            AnnotateTandemRepeats(
                name="annotate_tandem_repeats",
                param={
                    "$N": iSeq},
                **kwargs) for iSeq in self.lSeq]

        ParallelTaskCollection.__init__(self, self.tasks, **kwargs)

    def terminated(self):
        self.execution.returncode = 0
        gc3libs.log.info("\t\tSequencewiseParallelFlow.terminated")


def _get_requested_walltime_or_none(job):
    if isinstance(job, gc3libs.application.codeml.CodemlApplication):
        return job.requested_walltime.amount(hours)
    else:
        return None


# run script
if __name__ == '__main__':
    import tandem_repeat_annotation_workflow
    TandemRepeatAnnotationWorkflow().run()



import pytest

from tral.repeat import repeat
from tral.repeat import repeat_io
from tral.repeat.repeat_score import loglikelihood_gaps_starphylogeny_zipfian, \
    optimisation, loglikelihood_substitution, loglikelihood_substitutions_gaps


TEST_MSA = ['HPFGFV-------------AVPTKNP-DGTMNLMNWECAIPGKKGTPWEGGLFKLRMLFKDDYPS---SPPKCKFEPPLFHPNV',
            'YPSGTVCLsileedkdwrpAITIKQIlLGIQELLN-E---PNIQ-DPAQAEAYTIYCQNRVEYEKrvrAQAK-KFAP-------']
TEST_BEGIN = 0


@pytest.mark.notfixed
@pytest.mark.no_external_software_required
def test_repeat_score():

    # MAKE THIS TEST A TEST!
    Q, eqFreq, alphabet = repeat_io.loadModel()
    indelRatePerSite = 0.001

    myTR = repeat.Repeat(msa=TEST_MSA, begin=TEST_BEGIN)
    myTR.deleteInsertionColumns()

    print(loglikelihood_gaps_starphylogeny_zipfian(t=1, tandem_repeat=myTR))
    print(optimisation(function=loglikelihood_substitution, args=[Q, eqFreq, alphabet, myTR]))
    print(optimisation(function=loglikelihood_substitutions_gaps,
                       args=[[Q, eqFreq, alphabet, myTR], [myTR, indelRatePerSite]]))

    assert 1 == 2



import os
import pytest

from tral.sequence import sequence_io

TEST_FAA_FILE = "HIV-1_388796.faa"
TEST_SEQ_1 = "MKVMGIQRNCQQWWIWGILGFWMLMICNGMGNLWVTVYYGVPVWKDASPTLFCASDAKAYDTE" \
    "VHNVWGTFACVPTDPSPQELGLENVTENFNMWKNDMVEQMHQDIISLWDQGLKPCVKLTPLCVTLNCNAIKN" \
    "NTKVTNNSINSANDEMKNCSFNITTELRDKKRKAYALFYKLDIVPLNNGSTDYRLINCNTSTITQACPKVSL" \
    "DPIPIHYCAPAGYAILKCRDKTFTGTGPCHNVSTVQCTHGIKPVVSTQLLLNGSIAEGETIIRFENLTNNAK" \
    "IIIVQLNESVEITCTRPSNNTRESIRIGPGQTFYATGDIIGDIRQAHCNISEEKWNKTLQKVKEKLQKHFPN" \
    "KTIEFKPSSGGDLEITTHSFNCGGEFFYCNTSNLFNSTKLELFNSSTNLNITLQCRIKQIINMWQGVGRAMY" \
    "APPIEGIIMCRSNITGLLLTRDGAKEPHSTKEIFRPEGGDMRDNWRSELYKYKVVEIKPLGVAPTKPKRRVV" \
    "EREKRAALGALFLGFLGAAGSTMGAASITLTVQARQLLSGIVQQQSNLLKAIEAQQHMLQLTVWGIKQLQTR" \
    "VLAIERHLRDQQLLGIWGCSGKLICTTAVPWNSSWSNKSQEEIWDNMTWMQWDREISNYTDIIYNLLEVSQN" \
    "QQDKNEKDLLALDKWENLWNWFNITNWLWYIKIFIMIVGGVIGLRIIFAVLSIVNRVRQGYSPLSFQTLIPH" \
    "PRGPDRLGGIEEEGGEQGRDRSIRLVNGFLAIFWDDLRSLCLFSYHRLRDLILIAARTVELLGRSSLKGLQR" \
    "GWETLKYLGSLVQYWGLELKKSAINLLNTTAIVVGEGTDRFIELIQRIWRAFCNIPRRIRQGLEAALQ"
TEST_SEQ_1_ID = "sp|Q75008|ENV_HV1ET"


@pytest.fixture
def path():
    """Return the path to the test data files.
    """
    return os.path.dirname(os.path.abspath(__file__))


@pytest.mark.no_external_software_required
def test_read_fasta(path):

    test_lSeq = list(sequence_io.read_fasta(os.path.join(path, TEST_FAA_FILE)))
    assert len(test_lSeq) == 9
    assert test_lSeq[0] == (TEST_SEQ_1, TEST_SEQ_1_ID)



# (C) 2015 Elke Schaper


"""
    :synopsis: Alignment of tandem repeat units.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
    .. moduleauthor:: Paulina Näf
"""

from tral import configuration
import logging
import os
import subprocess
import tempfile

log = logging.getLogger(__name__)

CONFIG_GENERAL = configuration.Configuration.instance().config
REPEAT_CONFIG = CONFIG_GENERAL["repeat"]


def realign_repeat(my_msa, realignment='mafft', sequence_type='AA', rate_distribution=REPEAT_CONFIG['castor_parameter']['rate_distribution'], user_path=None):

    """ Realignment of a repeat MSA using mafft or proPIP

    A good multiple sequence alignment (MSA) is an important description of a
    tandem repeat (TR) and can be crucial to distinguish a true positive from
    a false positive TR.
    This can be realised using a proper MSA estimation algorithm. Currently,
    Mafft and proPIP can be used to realign a TR.

    For proPIP the rate distribution for indels can be either constant or gamma distributed
        - REPEAT_CONFIG['castor_parameter']['rate_distribution'] == 'constant': no gamma distribution is used for the proPIP algorithm
        - REPEAT_CONFIG['castor_parameter']['rate_distribution'] == 'gamma': gamma distribution (n=6, alpha=0.5)

    Args:
        my_msa (list of strings): List of sequences (str)
        realignment (str): Either "mafft" or "proPIP"
        sequence_type (str): Either "AA" or "DNA"
        user_path (str): copy alignment files to user defined path

    Returns:
        my_msa realigned (list of strings)

    .. todo:: 
        - decide what happens if branch length of tree is getting zero
    """

    realignment_types = ['mafft', 'proPIP', None]
    if realignment not in realignment_types:
        raise ValueError("Invalid realignment option. Expected one of: {}".format(realignment_types))

    if not all(isinstance(s, str) for s in my_msa):
        raise ValueError("Invalid MSA input.")

    if realignment == "proPIP":
        # remove columns that only contains gaps to avoid errors
        my_msa = remove_gaps(my_msa)

        # do try to realign if no gap is present in MSA
        if not any(['-' in s for s in my_msa]):
            print("\nproPIP needs at least one gap in the MSA to realign.\n" +
                  "No realignment will be performed.\n")
            return my_msa

    if realignment is None:
        print("\nNo realignment was performed.")
        return my_msa

    # Create temporary working directory
    working_dir = tempfile.mkdtemp()
    log.debug("evolvedTR: Created temp directory: %s", working_dir)

    # Save my_TR to temp directory:
    msa_file = os.path.join(working_dir, 'initial_msa.faa')
    with open(msa_file, 'w') as msa_filehandle:
        for i, iMSA in enumerate(my_msa):
            msa_filehandle.write('>{0}\n{1}\n'.format(i + 1, iMSA))

    if realignment == 'mafft':
        # Run Mafft
        # See http://mafft.cbrc.jp/alignment/software/manual/manual.html for choice of options.
        # The mafft result is in stdout. Check: Do you need to capture or
        # redirect the stderr?
        p = subprocess.Popen([REPEAT_CONFIG['ginsi'],
                              "--anysymbol",
                              "--quiet",
                              msa_file],
                             stdout=subprocess.PIPE)
        mafft_output = [line.decode('utf8').rstrip() for line in p.stdout]
        msa = []
        label = []
        for i, line in enumerate(mafft_output):
            if line[0] == '>':
                label.append(int(line[1:]))
            elif mafft_output[i - 1][0] == '>':
                msa.append(line)
            else:
                msa[-1] += line
        # use original order of msa
        msa_sorted = [x for _, x in sorted(zip(label, msa))]

        log.debug('\n'.join(msa_sorted))
        p.wait()
        try:
            return msa
        except:
            error_note = (
                "Mafft could not successfully run the realignment for: "
                "\n".join(my_msa))
            logging.error(error_note)
            return None

    elif realignment == 'proPIP':
        # Run Castor (with integrated aligner) (https://github.com/acg-team/castor_aligner)

        # log messages of castor to stderr instead of logfiles
        os.environ["GLOG_logtostderr"] = "1"

        if sequence_type == "AA":
            alphabet = "Protein"
            substitution_model = "LG08"
        elif sequence_type == "DNA":
            alphabet = "DNA"
            substitution_model = "HKY85"
        else:
            raise ValueError("Sequence type is not known.")

        if rate_distribution == 'constant':
            rate_distribution = 'Constant'
        elif rate_distribution == 'gamma':
            rate_distribution = 'Gamma(n=6,alpha=0.5)'
        else:
            raise ValueError("Rate distribution parameter not known.")

        tree_initial = os.path.join(working_dir, "tree_initial.nwk")
        tree = os.path.join(working_dir, "tree.nwk")
        msa_realigned = os.path.join(working_dir, "msa_realigned.faa")
        paramsfile_tree = os.path.join(working_dir, 'params_tree.txt')
        paramsfile_alignment = os.path.join(working_dir, 'params_alignment.txt')
        estimates = os.path.join(working_dir, 'estimates.json')

        ####################################
        # Create an initial tree
        ####################################

        init_tree = "distance"
        # Castor cannot create trees for less than four sequences

        if len([1 for line in open(msa_file) if line.startswith(">")]) == 2:
            print("For two units which have to be aligned an arbritary starting tree will be given.")
            tree_string = "(1:0.1,2:0.1);"
            init_tree = "user"
            with open(tree_initial, 'w') as treefile:
                treefile.write(tree_string)

        # For three units an arbritary initial tree is used for the alignment
        if len([1 for line in open(msa_file) if line.startswith(">")]) == 3:
            print("For three units which have to be aligned an arbritary starting tree will be given.")
            tree_string = "((1:0.1,3:0.1):0.1,2:0.1);"
            init_tree = "user"
            with open(tree_initial, 'w') as treefile:
                treefile.write(tree_string)
        # For more than tree units an initial tree will be estimated from the previous alignment

        # create parameter file to create or optimize tree with castor
        parameters_tree = ["analysis_name=tree_optimization",
                           "alphabet={}".format(alphabet),
                           "alignment=false",
                           "input.sequence.file={}".format(msa_file),
                           "input.sequence.sites_to_use=all",
                           "init.tree={}".format(init_tree),
                           "input.tree.file={}".format(tree_initial),
                           "init.distance.method=bionj",
                           "model=PIP(model={}(initFreqs=observed),initFreqs=observed)".format(substitution_model),
                           "rate_distribution=Constant",  # creating tree alwas with constant rates to estimate lamda and mu from data
                           "optimization=D-BFGS(derivatives=BFGS)",
                           "optimization.max_number_f_eval=500",
                           "optimization.tolerance=0.001",
                           "optimization.final=bfgs",
                           "optimization.topology=true",
                           "optimization.topology.algorithm=Swap(coverage=nnr-search,starting_nodes=Hillclimbing(n=4),max_cycles=50,tolerance=0.01,brlen_optimisation=Brent,threads=10)",
                           "output.tree.file={}".format(tree),
                           "output.tree.format=Newick",
                           "output.estimates.format=json",
                           "output.estimates.file={}".format(estimates),
                           "support=none"]
        try:
            with open(paramsfile_tree, 'w') as params:
                for parameter in parameters_tree:
                    params.write(parameter + '\n')
        except:
            print("A problem occurred while trying to write alignment parameters to txt file")

        # run castor for tree initialization
        try:
            castor_tree_initialization = subprocess.Popen([REPEAT_CONFIG['Castor'],
                                                           "params={}".format(paramsfile_tree)],
                                                          stdout=subprocess.PIPE,
                                                          stderr=subprocess.PIPE)
            print("Tree initialization with Castor ....")
            castor_tree_initialization.wait()

        except FileNotFoundError:
            error_note = (
                "ProPIP could not be reached.\n" +
                "Is Castor installed properly and is the path defined correctly in config.ini in the data directory?\n")
            logging.error(error_note)
            raise Exception("Sorry, creating a tree for the alignment went wrong.")

        ###################
        # Use proPIP algorithm to align TR units with inferred tree
        ###################

        # create file with unaligned sequences
        unaligned_sequences = os.path.join(working_dir, "sequences.faa")
        try:
            with open(msa_file, 'r') as infile, open(unaligned_sequences, 'w') as outfile:
                temp = infile.read().replace("-", "")
                outfile.write(temp)
        except:
            print("A problem occurred reading initial alignment file.")

        # parse estimates file to use calculated indel parameters
        try:
            import json
            with open(estimates) as estimates_file:
                est = json.load(estimates_file)
                mu_estimated = est['Model']['PIP']['mu']
                lambda_estimated = est['Model']['PIP']['lambda']
                indel_parameters = ',lambda={},mu={}'.format(lambda_estimated, mu_estimated)
        except FileNotFoundError as error:
            print('\nNo file with parameter estimates found.')
            print('Probably it went something wrong when trying to infer or optimise the phylogenetic tree.')
            logging.error(error)
            raise

        # TODO: proPIP cannot produce an alignment when a branchlength is zero.
        #       Need to catch this before an error is thrown
        #       Maybe replace zero with a very small number?
        #       Is this scenario even possible after tree calculation with castor?

        # create parameter file for alignment with proPIP
        parameters_alignment = ["analysis_name=aligner",
                                "model_description={}+PIP".format(substitution_model),
                                "alphabet={}".format(alphabet),
                                "alignment=true",
                                "alignment.version=ram",
                                "input.sequence.file={}".format(unaligned_sequences),
                                "input.sequence.sites_to_use=all",
                                "init.tree=user",
                                "init.distance.method=bionj",
                                "input.tree.file={}".format(tree),
                                "model=PIP(model={}{})".format(substitution_model, indel_parameters),
                                "rate_distribution={}".format(rate_distribution),
                                "optimization=None",
                                "output.msa.file={}".format(msa_realigned),
                                "support=none"]
        try:
            with open(paramsfile_alignment, 'w') as params:
                for parameter in parameters_alignment:
                    params.write(parameter + '\n')
        except:
            print("A problem occurred while trying to write alignment parameters to txt file")

        # run alignment with propip
        try:
            proPIP_alignment = subprocess.Popen([REPEAT_CONFIG['Castor'],
                                                "params={}".format(paramsfile_alignment)],
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.PIPE)
            print("Realignment of MSA with proPIP .... ")
            proPIP_alignment.wait()
        except FileNotFoundError:
            error_note = (
                "ProPIP algorithm could not be reached.\n" +
                "Is Castor (inkl. proPIP aligner) installed properly and is the path defined in config.ini in the data directory?\n")
            logging.error(error_note)
            return

        # read msa and return as sorted list
        try:
            # The created alignment file has "initial" included into the name because in future the tool should be able to realign
            with open(os.path.join(working_dir, "msa_realigned.initial.faa"), "r") as f:
                realigned = f.readlines()
        except FileNotFoundError:
            try:
                with open(msa_realigned, "r") as f:
                    realigned = f.readlines()
            except FileNotFoundError:
                error_note = (
                    "ProPIP could not successfully be used for the realignment of:\n" +
                    "\n".join(my_msa))
                logging.error(error_note)
                return
        msa = []
        label = []
        for i, line in enumerate(realigned):
            if line[0] == '>':
                label.append(int(line[1:-1]))
            elif realigned[i - 1][0] == '>':
                msa.append(line[:-1])
            else:
                msa[-1] += line[:-1]
        # use original order of msa
        msa_sorted = [x for _, x in sorted(zip(label, msa))]
        log.debug('\n'.join(msa_sorted))

        ## copy alignment files to user defined path
        if user_path:
            if not os.path.exists(os.path.dirname(user_path)):
                os.makedirs(os.path.dirname(user_path))
            import shutil
            mafft_path = user_path.split("proPIP")[0] + "mafft"
            with open(user_path, 'w') as out_msa:
                for i in range(len(msa_sorted)):
                    out_msa.write(">{}\n".format(i + 1))
                    out_msa.write("{}\n".format(msa_sorted[i]))
            shutil.copy(msa_file, mafft_path)

        try:
            return msa_sorted
        except:
            raise RuntimeError(
                "ProPIP could not successfully be used for the realignment of:\n" +
                "\n".join(my_msa))
    else:
        raise ValueError(
            'Currently, the aligner {} is not implemented.'.format(realignment))


def remove_char(str, n):
    """ helper function to remove a character in a sequence
    """
    first_part = str[:n]
    last_part = str[n + 1:]
    return first_part + last_part


def remove_gaps(msa):
    """ Remove columns in a MSA which contain only gaps.
    Do not use MSA longer than 3000 characters.

    Returns the MSA as a list without only-gap-containing columns.

    Args:
        msa (list of strings): List of sequences (str)

    Returns:
        msa (list of strings) with removed colums in case they only contain gaps
    """

    if not all(isinstance(s, str) for s in msa):
        raise ValueError("Invalid MSA input.")

    i = 3000  # The MSA should have at maximum 3000 characters

    if not all(len(s) < i for s in msa):
        raise ValueError("Length of sequences in MSA need to be below {}.".format(i))

    while i >= len(msa[0]):
        i -= 1

    while i <= len(msa[0]) - 1 and not i < 0:
        column = []
        for seq in msa:
            column.append(seq[i])
        # print(column)
        if all([char == '-' for char in column]):
            col = 0
            for seq in msa:
                seq = remove_char(seq, i)
                msa[col] = seq
                col += 1
        i -= 1

    return msa



import pytest

from tral.hmm.hmm import HMM
from tral.hmm.hmm_viterbi import hmm_path_to_non_aligned_tandem_repeat_units, \
    hmm_path_to_maximal_complete_tandem_repeat_units
from tral.repeat import repeat
from tral.repeat import repeat_align


@pytest.mark.no_external_software_required
def test_hmm_path_to_non_aligned_tandem_repeat_units():

    TEST = {"Double": ["AAAA", ["M1", "M2", "M1", "M2"], 2, ["AA", "AA"]],
            "Double_converted": ["ABCD", ["M2", "M1", "M2", "M1"], 2, ["AB", "CD"]],
            "Double_converted_complex": ["NNAAAACC",
                                         ["N", "N", "M2", "M1", "M2", "M1", "C", "C"],
                                         2,
                                         ["AA", "AA"]],
            "Double_converted_complex_insertions": ["NNAIAAACC",
                                                    ["N", "N", "M2", "I2", "M1", "M2", "M1", "C", "C"],
                                                    2,
                                                    ["AIA", "AA"]],
            "Single": ["AAAA", ["M1", "M1", "M1", "M1"], 1, ["A", "A", "A", "A"]],
            "Single_Complex": ["NNAAIIIAA",
                               ["N", "N", "M1", "M1", "I1", "I1", "I1", "M1", "M1"],
                               1,
                               ["A", "AIII", "A", "A"]],
            "Long": ["GYRADKLADKLADKL",
                     ["N", "N", "N", "M1", "M2", "M3", "M4", "M1", "M2", "M3", "M4", "M1", "M2", "M3", "M4"],
                     4,
                     ["ADKL", "ADKL", "ADKL"]],
            }
    for test, p in TEST.items():
        test_repeat_msa = hmm_path_to_non_aligned_tandem_repeat_units(sequence=p[0], path=p[1], l_effective=p[2])
        assert test_repeat_msa == p[3]


def test_viterbi():

    # {Test_name: [Original_TR_MSA, Sequence, Viterbi_path, Refined_TR_MSA], ... }
    TEST = {"Single": [["A", "A", "A"], "AAAAAA", ["M1", "M1", "M1", "M1", "M1", "M1"], ["A", "A", "A", "A", "A", "A"]],
            "Double": [["AA", "AA"], "AAAAAA", ["M1", "M2", "M1", "M2", "M1", "M2"], ["AA", "AA", "AA"]],
            "Long": [["ADKL", "ADKL"],
                     "GYRADKLADKLADKL",
                     ["N", "N", "N", "M1", "M2", "M3", "M4", "M1", "M2", "M3", "M4", "M1", "M2", "M3", "M4"],
                     ["ADKL", "ADKL", "ADKL"]]
            }

    for test, p in TEST.items():
        test_repeat = repeat.Repeat(msa=p[0])
        test_hmm = HMM.create(input_format="repeat", repeat=test_repeat)

        for iHMM in [test_hmm]:
            # Detect TRs on self.seq with hmm using the Viterbi algorithm.
            most_likely_path = iHMM.viterbi(p[1])
            assert type(most_likely_path) == list
            assert most_likely_path == p[2]

            unaligned_msa = hmm_path_to_non_aligned_tandem_repeat_units(p[1], most_likely_path, iHMM.l_effective)
            assert unaligned_msa == p[3]

            aligned_msa = repeat_align.realign_repeat(unaligned_msa)
            assert aligned_msa == p[3]

# The following test functions were copy-pasted from hmm_viterbi.py. They need to be either
# strongly adapted, or - if not useful - erased.


@pytest.mark.notfixed
@pytest.mark.no_external_software_required
def test_conversion_multiple():
    """ Test <hmm_path_to_maximal_complete_tandem_repeat_units>.
        However, no positive set is defined right now.
    """
    assert 0, "Test not fixed"
    l_effective = 3
    lPaths = [['N', 'N', 'N', 'M0', 'M1', 'M2', 'M0', 'I1', 'M1', 'M2', 'M0', 'M1', 'M2', 'N', 'N', 'N'],
              ['N', 'N', 'N', 'M1', 'M2', 'M0', 'M1', 'M2', 'M0', 'M1', 'M2', 'M0', 'N', 'N', 'N'],
              ['N', 'N', 'N', 'M2', 'M0', 'M1', 'M2', 'M0', 'M1', 'M2', 'M0', 'M1', 'N', 'N', 'N']]
    lSequence = ['XXXABCADBCABCXXX', 'XXXBCABCABCAXXX', 'XXXCABCABCABXXX']
    lMSA = hmm_path_to_maximal_complete_tandem_repeat_units(lSequence, lPaths, l_effective)
    assert lMSA is not None


@pytest.mark.notfixed
@pytest.mark.no_external_software_required
def test_create_viterbi():
    """ This test needs to be fixed, the HMM cannot be initialised in this way.
    """
    assert 0, "Test not fixed"
    my_hmm = HMM()
    my_hmm.states = ['H', 'F']

    # Initialisation
    my_hmm.p_0 = {"H": 0.6, "F": 0.4}

    # Feed Values to p_t
    #my_hmm.p_t["START"] = {"H": 0.6, "F": 0.4}
    my_hmm.p_t["H"] = {"H": 0.7, "F": 0.3}
    my_hmm.p_t["F"] = {"H": 0.4, "F": 0.6}

    # emissions
    my_hmm.emissions = ["D", "C", "N"]

    # emission probabilities
    my_hmm.p_e['H'] = {"D": 0.1, "C": 0.4, "N": 0.5}
    my_hmm.p_e['F'] = {"D": 0.6, "C": 0.3, "N": 0.1}

    my_viterbi = Viterbi(my_hmm, "NCD")
    if my_viterbi.viterbi() == "HHF":
        print("Test SUCCESSFUL")
    else:
        print("Test FAILED")


@pytest.mark.notfixed
def test_init_with_repeat():
    """ This test needs to be fixed.
    """
    assert 0, "Test not fixed"
    my_TR = repeat.Repeat(begin=0, msa=['A-G', 'ACG', 'ACG'], sequence_type='DNA')
    my_hmm = HMM.create(repeat=my_TR)
    from . import sequence
    my_sequence = sequence.Sequence()
    my_viterbi = Viterbi(my_hmm, my_sequence.sequence)
    v = my_viterbi.viterbi()
    "".join(v)



# import pytest

from tral.repeat import repeat_align

TEST_MSA = ["MGKGYL---------------------------------------ALCSYNCKEA-INILSHLPSHHYN",
            "TG--------------------------------------------------------------WVLCQ",
            "IGRAYF---------------------------------------ELSEYMQAER-IFSEVRRIENYRV",
            "EGMEIYSTTLWHLQK------------------------------DVALSVLSKDLTDMDKNSPEAWCA",
            "AGNCFS---------LQREH-------------------------DIAIKFFQRA-IQVDPNYAYAYTL",
            "LGHEFV--------------LTEEL--------------------DKALACFRNA-IRVNPRHYNAWYG",
            "LGMIYY-------------------KQEKF---------------SLAEMHFQKA-LDINPQSSVLLCH",
            "IGVVQH------------------------ALKKS----------EKALDTLNKA-IVIDPKNPLCKFH",
            "RASVLF-----------------------------ANEKY-----KSALQELEEL-KQIVPKESLVYFL",
            "IGKVYK----------------------------------KLGQTHLALMNFSWA-MDLDPKGAN----"]
TEST_REALIGNED_MSA = ['MGKGYLALC----SYNC-----KEAI-NILSHLPSHHYN',
                      'TGWVL--------------------------------CQ',
                      'IGRAYFELS----EYMQAERIFSEVR-RIEN-----YRV',
                      'EGMEIYSTTLWHLQKDVALSVLSKDLTDMDKNSPEAWCA',
                      'AGNCFSLQR----EHDIAIKFFQRAI-QVDPNYAYAYTL',
                      'LGHEFVLTE----ELDKALACFRNAI-RVNPRHYNAWYG',
                      'LGMIYYKQE----KFSLAEMHFQKAL-DINPQSSVLLCH',
                      'IGVVQHALK----KSEKALDTLNKAI-VIDPKNPLCKFH',
                      'RASVLFANE----KYKSALQELEELK-QIVPKESLVYFL',
                      'IGKVYKKLG----QTHLALMNFSWAM-DLDPKGA----N']


def test_repeat_alignment():
    ''' Test a realignment using Mafft's ginsi.'''

    REALIGNED_MSA = repeat_align.realign_repeat(my_msa=TEST_MSA)

    assert isinstance(REALIGNED_MSA, list)

    # Attention: If ginsi changes, the alignment will also change...
    assert REALIGNED_MSA == TEST_REALIGNED_MSA



# (C) 2015 Elke Schaper

"""
    :synopsis: Input/output for tandem repeats.

    .. moduleauthor:: Elke Schaper <elke.schaper@isb-sib.ch>
"""

import Bio.Seq
import random
import logging
import itertools
import os
import re
import shutil
import tempfile
import subprocess
import numpy as np
from tral.repeat import repeat
from tral import configuration

from tral.paths import DATA_DIR
CONFIG_GENERAL = configuration.Configuration.instance().config
REPEAT_CONFIG = CONFIG_GENERAL["repeat"]

LOG = logging.getLogger(__name__)


###################### SAVE REPEAT #######################################

def save_repeat_fasta(tandem_repeats, file):
    """ save multiple <tandem_repeats> in Fasta format in specified <file>

        At current, only one TR per sequence can be defined, as the identifiers
        in the dict <tandem_repeats> must be unique.

        Parameters: Dict of tandem repeats and identifiers.
            e.g. {'ENSP00012': msa1, 'ENSP00013': msa2}

            >ID
            GHKI
            GHKI
            GH--
    """

    with open(file, 'w', newline='\n') as f:
        for identifier, msa in tandem_repeats.items():
            f.write(">{0}\n".format(identifier))
            f.write("\n".join(msa) + "\n\n")


def save_repeat_stockholm(tandem_repeat, file):
    """ save <tandem_repeat> in STOCKHOLM format in specified <file>

        Parameters: Tandem repeat MSA.
            e.g. ["ACDEF-", "ACCDEF"]

            More information in
            ftp://selab.janelia.org/pub/software/hmmer3/3.0/Userguide.pdf
            STOCKHOLM Format Example:

            # STOCKHOLM 1.0
            seq1 ACDEF...GHIKL
            seq2 ACDEF...GHIKL
            seq3 ...EFMNRGHIKL

            seq1 MNPQTVWY
            seq2 MNPQTVWY
            seq3 MNPQT...
            //

    """

    with open(file, 'w', newline='\n') as fh:
        fh.write("# STOCKHOLM 1.0\n")
        for i, iMSA in enumerate(tandem_repeat):
            fh.write("{} {}\n".format(str(i), iMSA))
        fh.write("//")


def save_repeat_treks(tandem_repeats, file):
    """ Save multiple `tandem_repeats` in T-REKS format in specified `file`

        At current, only one TR per sequence can be defined, as the identifiers
        in the dict <tandem_repeats> must be unique.

        Parameters: Dict of tandem repeats and identifiers.
            e.g. {'ENSP00012': [msa1, begin1], 'ENSP00013': [msa2, begin2]}

            T-REKS format example:

            >a
            Length: 3 residues - nb: 3  from  1 to 10 - Psim:0.8076923076923077 region Length:42
            GHKI
            GHKI
            GH--
            **********************

            Length: 3 residues - nb: 2  from  21 to 27 - Psim:0.7857142857142857 region Length:22
            GHKI
            GH--
            **********************

            >b
            Length: 3 residues - nb: 3  from  1 to 10 - Psim:0.8095238095238095 region Length:20
            UIR
            UIR
            UIR
    """

    with open(file, 'w', newline='\n') as fh:
        for identifier, info in tandem_repeats.items():
            fh.write(">{0}\n".format(identifier))
            fh.write("Length: from {0} to\n".format(str(info[1])))
            fh.write("\n".join(info[0]) + "\n" + "*" * 22 + "\n\n")


############################## READ REPEAT SEQUENCE ######################


def read_fasta(seq_filename, sequence_type='AA'):
    """ Read repeat from file in fasta format.

    Read repeat from file in fasta format.

    Args:
        seq_filename (str):  Path to the repeats containing fasta file
        sequence_type (str): Either "AA" or "DNA"

    Returns:
        (list of Repeat): A list of Repeat instances.
    """

    pat_start = re.compile(r">(.*)")
    pat_repeat_unit = re.compile(r"([\w\.\-]+)")

    # Our possible parser states:
    #
    # 1: searching for sequence name
    # 2: searching for repeat units

    repeats = {}
    state = 1
    with open(seq_filename, "rt") as infile:

        for i, line in enumerate(infile):
            LOG.debug("Line {0}: {1}".format(i, line[0:-1]))
            if 1 == state:
                match = pat_start.match(line)
                if match:
                    LOG.debug(" * (1->2) Found start")
                    LOG.debug("Start: %s", match.group(1))
                    name = match.group(1)
                    repeats[name] = []
                    state = 2

            elif 2 == state:
                match = pat_repeat_unit.match(line)
                if match:
                    LOG.debug(" * (2->2) Found Repeat unit")
                    LOG.debug("Repeat Unit: %s", match.group(1))
                    repeat_unit = match.group(1)
                    repeats[name].append(repeat_unit.replace(".", "-").upper())
                else:
                    LOG.debug(" * (2->1) Found NO further Repeat unit")
                    state = 1

    for i_name, iR in repeats.items():
        yield iR, sequence_type


################################### SIMULATE SEQUENCE ####################

def evolved_tandem_repeats(l, n, n_samples, sequence_type, job_id='job_id',
                           mutation_rate=50, tree='star',
                           indel_rate_per_site=False,
                           return_type='repeat'):
    """ Simulate evolved sequences with ALF.

    Simulate evolved sequences with ALF:
    Dalquen, D. A., Anisimova, M., Gonnet, G. H. & Dessimoz, C.
    ALF--a simulation framework for genome evolution. Molecular Biology and Evolution 29,
    1115–1123 (2012).

    Args:
        l (int): The length of the repeat unit
        n (int): The number of repeat units in the tandem repeat
        n_samples (int):  The number of samples
        sequence_type (str): Either "AA" or "DNA"
        job_id (str): A tag for files produces with ALF, and result files.
        mutation_rate (float): The mutation rate.
        tree (str): The type of tree, e.g. "star" or "birthdeath"
        indel_rate_per_site (int or False): The indel rate per site.
        return_type (str): Either "repeat" or "list"

        sequence_length (int): The total length of the simulated sequence

    Returns:
        Return type depends on ``return_type``: ``Repeat`` or ``Bio.Seq.Seq`` instance.
    """

    runfile_template = os.path.join(DATA_DIR, "ALF", "template.drw")
    alf_exec = REPEAT_CONFIG['alfsim']
    # create temporary directory
    working_dir = tempfile.mkdtemp()
    LOG.debug("evolvedTR: Created tempfile: %s", working_dir)

    # create working dir
    if not os.path.isdir(working_dir):
        os.makedirs(working_dir)

    # Copy template file and append job specific info
    runfilename = "alf.drw"
    shutil.copyfile(runfile_template, os.path.join(working_dir, runfilename))

    with open(os.path.join(working_dir, runfilename), "a") as runfile:
        if indel_rate_per_site:
            # insertion rate per site and PAM. E.g. for PAM=40 expect
            # aaGainRate*40 insertions per site.
            runfile.write(
                "aaGainRate := " + str(indel_rate_per_site / mutation_rate) + ";\n")
            runfile.write(
                "aaLossRate := " + str(indel_rate_per_site / mutation_rate) + ";\n")
            runfile.write("maxIndelLength := 50;\n")
            runfile.write("indelModel := 'ZIPF';\n")
            runfile.write("Z_c := 1.821;\n")
            runfile.write("DawgPlacement := true;\n")
        runfile.write("uuid := '" + job_id + "';\n")
        runfile.write("mname := '" + job_id + "';\n")
        runfile.write("protStart := " + str(n_samples) + ";\n")
        runfile.write("NSpecies := " + str(n) + ";\n")
        runfile.write("minGeneLength := " + str(l) + ";\n")
        runfile.write("mutRate := " + str(mutation_rate) + ";\n")
        runfile.write("wdir := '" + working_dir + "';\n")

        tree_length = n * mutation_rate
        # parameters concerning the species tree
        if tree == 'star':
            # BDTree, ToLSample, Custom
            runfile.write("treeType := 'Custom':\n")
            runfile.write(
                "tree := MakeStarTree(BirthDeathTree(0.1,0.1," + str(n) + ",10),mutRate):\n")
            runfile.write("treeLength:= %d ;\n" % tree_length)
        else:
            # BDTree, ToLSample, Custom
            runfile.write("treeType := 'BDTree':\n")
            # From the last discussion with Daniel, only the ration of birth to
            # death rate matters, if we scale tree to match pam distance
            runfile.write("birthRate := 0.01:\n")
            runfile.write("deathRate := 0.01:\n")
            # for BDTree: should resulting tree be ultrametric, e.g. all leaves
            # have same distance to origin?
            runfile.write("ultrametric := false:\n")
            runfile.write("treeLength:= %d ;\n" % tree_length)
            # DANIEL: Is a tree := missing?
        if tree not in {'star', 'birthdeath'}:
            LOG.warning(
                "evolvedTR: tree input %s not known, assuming birthdeath tree",
                tree)

        # parameters concerning the substitution models
        if sequence_type == 'AA':
            runfile.write(
                "substModels := [SubstitutionModel('CustomP', ['" +
                os.path.join(
                    DATA_DIR,
                    'ALF',
                    'lg.dat') +
                "'])];\n")
            # CHECK! DOES
            # substModels := [SubstitutionModel('LG')];
            # WORK?
        elif sequence_type == 'DNA':
            runfile.write(
                "substModels := [SubstitutionModel('TN93', [.3, .4, .7],[seq(0.25,4)], true)]:\n")
            # CHECK! DO THE EMPIRICAL PARAMETERS MAKE SENSE AT ALL?
            # compare to http://people.inf.ethz.ch/ddalquen/alf/ALF_manual.pdf
        else:  # CODONS
            runfile.write("substModels := [SubstitutionModel('CPAM')];\n")

    # Determine ALF MSA output file path
    if sequence_type == 'AA':
        infile = os.path.join(working_dir, job_id, "MSA", "MSA_all_aa.phy")
    elif sequence_type == 'DNA':
        infile = os.path.join(working_dir, job_id, "MSA", "MSA_all_dna.phy")
    else:  # CODONS
        infile = os.path.join(
            working_dir,
            job_id,
            "MSA",
            "MSA_all_codon.phy")

    for i in range(10):
        with open(os.path.join(working_dir, "out.txt"), "w") as outfile:
            alf_process = subprocess.Popen([alf_exec,
                                            runfilename],
                                           cwd=working_dir,
                                           stdout=outfile,
                                           stderr=outfile,
                                           close_fds=True)
            alf_process.wait()
        if os.path.isfile(infile):
            break
    else:
        LOG.error('ALFSIM was not able to produce simulated sequence.')
        LOG.error('Do you have ALFSIM installed and used the correct path in the configuration file?')
        return

    # shutil.rmtree('/cluster/home/infk/eschaper/spielwiese/')
    #shutil.copytree(working_dir, '/cluster/home/infk/eschaper/spielwiese/')

    """ Read in MSA data from ALF
        The following is a short parser for ALFsim output files
        yielding MSAs as lists of strings,
        based on a special flavour of  Felstein stein MSA files """

    # find a repeat unit
    pattern_start = re.compile(r"\d+ \d+")
    pattern_seq = re.compile(r"\S+[ ]+([A-Z\-]+)")

    # Our possible parser states:
    #
    # state 1: Find beginning of MSA (Felsenstein)
    # state 2: Find all repeat units

    # protein ::=
    #    \d \d
    #    \S/\S    repeatunits
    #

    state = 1
    with open(infile, "r") as infile:
        for i, line in enumerate(infile):
            LOG.debug("Line %d: %s", i, line[0:-1])

            if 1 == state:  # Find first repeat unit & save begin
                search = pattern_start.search(line)
                if search:
                    LOG.debug(" *(1->2) Found MSA start")
                    state = 2
                    msa = []

            elif 2 == state:  # Find all repeat units
                search = pattern_seq.search(line)
                if search:
                    LOG.debug(" *(2->2) Found another repeat unit")
                    msa.append(search.group(1))
                else:
                    LOG.debug(" *(2->1) repeat region finished, yielding.")
                    state = 1
                    # YIELD IF WE HAVE FOUND AT LEAST TWO REPEAT UNITS:
                    if len(msa) > 1:
                        if return_type == 'repeat':
                            yield repeat.Repeat(begin=0, msa=msa, sequence_type=sequence_type)
                        elif return_type == 'list':
                            yield msa
                        else:
                            LOG.debug("YIELD: %s",
                                      "".join(msa).replace('-', ''))
                            yield Bio.Seq.Seq("".join(msa).replace('-', ''), sequence_type)

    # delete temporary directory
    try:
        shutil.rmtree(working_dir)
    except OSError:
        pass


def random_sequence(n_samples, sequence_type='AA', return_type='repeat',
                    equilibrium_frequencies='human', l=0, n=0,
                    sequence_length=0):
    """ Simulate random sequence locally.

    Simulate random sequence locally.

    Args:
        n_samples (int):  The number of samples
        sequence_type (str): Either "AA" or "DNA"
        return_type (str): Either "repeat" or "list"
        equilibrium_frequencies (str): Only "human" option available at current
        l (int): The length of the repeat unit
        n (int): The number of repeat units in the tandem repeat
        sequence_length (int): The total length of the simulated sequence

    Returns:
        Return type depends on ``return_type``.
    """

    if sequence_length == 0 and (l == 0 or n == 0):
        LOG.error('The specified sequence_length or the product of l and n was'
                  ' set to 0 for random_sequence simulation')
    else:
        if sequence_length == 0:
            sequence_length = l * n

        if equilibrium_frequencies == 'human':
            file = os.path.join(DATA_DIR, 'Random', "_".join(
                [sequence_type, equilibrium_frequencies, '3']) + '.txt')

        with open(file, 'r') as f:
            a = f.readline()[:-1]
        b = [i.split(' ') for i in a.split('  ')]
        frequencies = {i[0]: int(i[1]) for i in b}
        alphabet = np.unique([i[0] for i in frequencies.keys()])

        for _ in range(n_samples):
            seed_int = random.randint(1, sum(frequencies.values()))
            for key, value in frequencies.items():
                seed_int -= value
                if seed_int <= 0:
                    seed = key
                    break

            dimer = [''.join(i) for i in itertools.product(alphabet, repeat=2)]
            third_letter_frequencies = {
                iD: {
                    iA: frequencies[
                        iD +
                        iA] for iA in alphabet} for iD in dimer}

            sequence = seed
            for _ in range(sequence_length - 3):
                next_int = random.randint(
                    1, sum(third_letter_frequencies[sequence[-2:]].values()))
                for key, value in third_letter_frequencies[
                        sequence[-2:]].items():
                    next_int -= value
                    if next_int <= 0:
                        sequence += key
                        break

            # return Repeat instances
            if return_type == 'repeat' and not l == 0 and not n == 0:
                yield [sequence[i * l:(i + 1) * l] for i in range(n)], sequence_type
            elif return_type == 'list':
                yield sequence
            else:  # return seqIO instances
                yield Bio.Seq.Seq(sequence, sequence_type)



import os

from tral.hmm.hmm import HMM
from tral.repeat import repeat

TEST_SEQUENCE = "GYEDEDEDRPFYALGLGKRPRTYSFGL"

TEST_REPEAT_MSA_DOUBLE = ["AA", "AA"]
TEST_HMM_STATES_DOUBLE = ["N", "M1", "I1", "M2", "I2", "C"]
TEST_HMM_P0_DOUBLE = {i: 0 for i in TEST_HMM_STATES_DOUBLE}


TEST_REPEAT_MSA_SINGLE = ["A", "A", "A"]
TEST_HMM_STATES_SINGLE = ["N", "M1", "I1", "C"]
TEST_HMM_P0_SINGLE = {i: 0 for i in TEST_HMM_STATES_SINGLE}


def test_create_HMM_from_Repeat():

    test_repeat = repeat.Repeat(msa=TEST_REPEAT_MSA_DOUBLE)
    test_hmm = HMM.create(input_format='repeat', repeat=test_repeat)

    assert test_hmm.l_effective == 2
    assert set(test_hmm.states) == set(TEST_HMM_STATES_DOUBLE)
    assert test_hmm.p_0 == TEST_HMM_P0_DOUBLE
    #assert test_hmm.p_t == TEST_HMM_P0_DOUBLE

    test_repeat = repeat.Repeat(msa=TEST_REPEAT_MSA_SINGLE)
    test_hmm = HMM.create(input_format='repeat', repeat=test_repeat)

    assert test_hmm.l_effective == 1
    assert test_hmm.states == TEST_HMM_STATES_SINGLE
    assert test_hmm.p_0 == TEST_HMM_P0_SINGLE


def test_create_HMM_from_DNA_Repeat():
    # no calculation of pvalue possible since there are no pvalue distribution files available

    test_repeat = repeat.Repeat(msa=TEST_REPEAT_MSA_DOUBLE, sequence_type="DNA",calc_pvalue=False)
    #test_parameters = {"hmmbuild":{"hmm_copy_path":"/Users/elkeschaper/Downloads", "hmm_copy_id":"maulwurf"}}
    #test_hmm = HMM.create(input_format = 'repeat', repeat = test_repeat, **test_parameters)
    test_hmm = HMM.create(input_format='repeat', repeat=test_repeat)

    assert set(test_hmm.hmmer['letters']) == set(["A", "C", "G", "T"])
    assert set(test_hmm.alphabet) == set(["A", "C", "G", "T"])

    assert test_hmm.l_effective == 2
    assert set(test_hmm.states) == set(TEST_HMM_STATES_DOUBLE)
    assert test_hmm.p_0 == TEST_HMM_P0_DOUBLE

    for iState, ip_e in test_hmm.p_e.items():
        assert set(list(ip_e.keys())) == set(["A", "C", "G", "T"])
    #assert test_hmm.p_t == TEST_HMM_P0_DOUBLE


def test_hmm_pickle(tmpdir):

    test_repeat = repeat.Repeat(msa=TEST_REPEAT_MSA_DOUBLE)
    test_hmm = HMM.create(input_format='repeat', repeat=test_repeat)

    test_pickle = os.path.join(tmpdir,"test.pickle")
    test_hmm.write(test_pickle, 'pickle')
    test_hmm_new = HMM.create(input_format='pickle', file=test_pickle)

    assert test_hmm.hmmer == test_hmm_new.hmmer
    assert test_hmm.alphabet == test_hmm_new.alphabet

    if os.path.exists(test_pickle):
        os.remove(test_pickle)


import os
import shutil
import sys
import logging

from setuptools import setup
from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.bdist_egg import bdist_egg as _bdist_egg

import subprocess

def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), "r") as f:
        return f.read()


# The previous --home option has been changed to --tral-home (0.3.7)
# Warn users about backwards incompatibility
# (`setup.py install` already has a --home option)
if "--home" in sys.argv:
    logging.warning("The option to set the TRAL data directory has been renamed from --home to --tral-home. "
                    "The --home option may be recognized instead to install in the user's home directory.")


class TralMixin(object):
    """Mixin to add additional parameters to setuptools commands
    """

    user_options = [
        ('tral-home=', None, 'Directory containing the `.tral` data directory (default to user home)'),
    ]

    def initialize_options(self):
        super().initialize_options()
        self.tral_home = os.path.expanduser("~")

    def finalize_options(self):
        super().finalize_options()
        if not os.path.exists(self.tral_home):
            raise ValueError("The argument supplied in --tral-home is not a valid path: {}".format(self.tral_home))

    def run(self):
        datadir = os.path.join(self.tral_home, ".tral")
        if os.path.exists(datadir):
            print("The TRAL configuration directory {} already exists. The "
                  "template configuration and datafiles are not copied to the "
                  "already existing directory at this step.".format(datadir))
        else:
            shutil.copytree("tral/tral_configuration", datadir)
            print("The TRAL configuration files and data files are now located in {}".format(datadir))

        try:
            # install's run() command ignores install_requires when called this way.
            # Instead, use do_egg_install if possible. https://stackoverflow.com/q/21915469/81658
            # super().do_egg_install()
            subprocess.call(["pip install -r requirements.txt --no-clean"], shell=True)
            _bdist_egg.run(self)
        except AttributeError:
            # fallback for develop
            super().run()


class InstallCommand(TralMixin, install):
    "Add TRAL options to setup.py install"
    user_options = getattr(install, 'user_options', []) + TralMixin.user_options


class DevelopCommand(TralMixin, develop):
    "Add TRAL options to setup.py develop"
    user_options = getattr(develop, 'user_options', []) + TralMixin.user_options


SCRIPTS1 = [os.path.join("tral", "examples", i) for i in ["example_workflow_MBE2014.py"]]
SCRIPTS2 = [os.path.join("tral", "examples", "workflow", i) for i in ["tandem_repeat_annotation_scripts.py",
                                                                      "tandem_repeat_annotation_workflow.py"]]
packages = ["tral", "tral.test", "tral.hmm", "tral.hmm.test", "tral.repeat", "tral.repeat.test",
            "tral.repeat_list", "tral.repeat_list.test", "tral.sequence", "tral.sequence.test"]

# Load the version number from tral/__init__.py
__version__ = "Undefined"
for line in open('tral/__init__.py'):
    if (line.startswith('__version__')):
        exec(line.strip())

setup(
    name="tral",
    version=__version__,
    author="Elke Schaper",
    author_email="elke.schaper@isb-sib.ch",
    packages=packages,
    scripts=SCRIPTS1 + SCRIPTS2,
    url="http://pypi.python.org/pypi/tral/",
    license="LICENSE.txt",
    description="Detect and evaluate tandem repeats in genomic sequence data.",
    long_description=read("README.rst"),
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Natural Language :: English",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "biopython >= 1.64",
        "configobj >= 5.0.6",
        "numpy >= 1.6.1",
        "scipy >=0.12.0",
    ],
    setup_requires=["pytest-runner"],
    tests_require=[
        "pytest >= 2.5.2",
    ],
    # Install with e.g. `python setup.py install tral[docs]`
    extras_require={
        'docs': [
            "docutils >= 0.11",
            "pypandoc >= 0.9.6",
            "Sphinx >= 1.2.2",
        ],
        'develop': [
            "flake8 >= 3.6",
            "flake8-colors",
            "tox >= 3.5",
            "pytest >= 2.5.2",
        ],
        'workflow': [
            "pyfaidx==0.4.7.1",  # Used by the workflow example for fasta indexing
        ]
    },
    # package_data: None-module files, which should still be distributed are mentioned here:
    package_data={"tral": ["tral_configuration/*.ini",
                           "tral_configuration/data/*",
                           "examples/*.py", "examples/data/*",
                           "examples/workflow/*.py", "examples/workflow/*.tsv",
                           "examples/workflow/*.ini", "examples/workflow/*.hmm",
                           "examples/workflow/*.fasta",
                           "examples/workflow/split_sequence_data/*.fasta"]},
    package_dir={"tral": "tral"},
    cmdclass={
        'install': InstallCommand,
        'develop': DevelopCommand,
        'bdist_egg': _bdist_egg,
    }
)
